import numpy as np


def mirrorAxis(picture, verticalFlip: bool = True, pos=0.5,  flip=False):
    """
    Mirror image along axis in given position.

    :param picture: array, 2d, 3d

    :param verticalFlip: bool
            True - Image will be mirrored up-down
            False - Image will be mirrored left-right

    :param pos: Float or Int
            Float - <0, 1> Axis position = number * dimention.
            Int - <0, MaxDimension> Image position

    :param flip: bool, flip other direction

    :return:

    """

    if len(picture.shape) == 3:
        h, w, c = picture.shape
    else:
        h, w = picture.shape

    if isinstance(pos, int):
        if verticalFlip:
            center = np.clip(pos, 0, h)
        else:
            center = np.clip(pos, 0, w)

    elif isinstance(pos, float):
        if verticalFlip:
            center = np.round(h * pos).astype(int)
        else:
            center = np.round(w * pos).astype(int)
    else:
        raise ValueError("Pos must be int or float")

    if verticalFlip:
        "Vertical mirror"
        if center == h or center == 0:
            "EDGE CASES"
            return np.flipud(picture)
        first = picture[:center, :]
        second = picture[center:, :]

    else:
        "Horizontal Mirror"
        if center == w or center == 0:
            "EDGE CASES"
            return np.fliplr(picture)
        first = picture[:, :center]
        second = picture[:, center:]

    " NORMAL MIRROR "
    if verticalFlip:
        if flip:
            first = np.flipud(second)
        else:
            second = np.flipud(first)

    else:
        if flip:
            first = np.fliplr(second)
        else:
            second = np.fliplr(first)

    axis = 0 if verticalFlip else 1
    combined = np.concatenate([first, second], axis=axis)

    return combined


__all__ = [
    'mirrorAxis',
]


if __name__ == "__main__":
    import cv2 as _cv2
    import os
    img = _cv2.imread(os.path.join(os.path.dirname(__file__), "cat.jpg"))

    count = 0
    imFlip = mirrorAxis(img, False, 0.3, False)

    "Loop checking"
    for pos in [0.2, 0, 400, 0.8]:
        for verticalFlip in [False, True]:
            for flip in [False, True]:
                print()
                imFlip = mirrorAxis(img, verticalFlip, pos, flip)
                # _cv2.imshow(str(count), imFlip)
                count += 1

    # mir1 = mirrorAxis(img, )
    # mir2 = mirrorAxis(img, flip=True)
    # _cv2.imshow("Orig", img)
    # _cv2.imshow("Smol1", mir1)
    # _cv2.imshow("Smol2", mir2)

    # _cv2.waitKey()



import pytest
from yasiu_image.filters import mirrorAxis

import numpy as np


@pytest.fixture()
def blank_image_flat():
    img = np.zeros((300, 300), dtype=np.uint8)
    return img


@pytest.fixture()
def blank_image_gray():
    img = np.zeros((300, 300, 1), dtype=np.uint8)
    return img


@pytest.fixture()
def blank_image():
    img = np.zeros((300, 300, 3), dtype=np.uint8)
    return img


tests_params = [
        (val, ax, flag)
        for val in [*np.linspace(0, 1, 15), *range(290, 310), *range(-15, 20)]
        for ax in [0, 1]
        for flag in [False, True]
]


# @pytest.fixture()
# def all_img_types():
#     image_params = [blank_image(), blank_image_gray(), blank_image_flat()]
#     return image_params


@pytest.mark.parametrize('val,ax,flag', tests_params)
@pytest.mark.parametrize('image', ['blank_image', 'blank_image_flat', 'blank_image_gray'])
def test_1(image, val, ax, flag, request):
    image = request.getfixturevalue(image)
    # print(image)
    # print(image.shape)
    # raise ValueError
    ret = mirrorAxis(image, pos=val, verticalAxis=ax, flip=flag)
    assert ret.shape == image.shape, "Shape must be true"
    # assert (image == ret).all(), "Returned diff image"



import numpy as _np

from PIL import Image as _Image
import cv2 as _cv2


# from yasiu_image.filters import mirror


def read_gif_frames(path):
    img = _Image.open(path, )
    ind = 0
    # sequence = []
    # img = img.convert("RGBA")
    img.seek(0)
    # fr = np.array(img, dtype=np.uint8)
    while True:
        fr = _np.array(img, dtype=_np.uint8).copy()
        # print(f"Read shape: {fr.shape}")
        # sequence.append(fr)
        yield fr

        ind += 1
        try:
            img.seek(ind)
        except EOFError:
            # print(f"Breaking at: {ind}")
            break


def read_webp_frames(path):
    img = _Image.open(path)
    ind = 0

    img.seek(0)
    # fr = np.array(img, dtype=np.uint8)
    while True:
        fr = _np.array(img, dtype=_np.uint8).copy()
        yield fr

        ind += 1
        try:
            img.seek(ind)
        except EOFError:
            # print(f"Breaking at: {ind}")
            break


def save_image_list_to_gif(frames, exp_path, use_rgba=False, duration=40, quality=100, disposal=2):
    """

    Args:
        frames:
        exp_path: path to export file with ".gif" ending
        use_rgba: bool,
        duration: int, default 40, [ms], preferable in range <20, 80>
        quality: 100
        disposal: int, default 2 = clear

    Returns:

    """

    if not (exp_path.endswith("gif") or exp_path.endswith("GIF")):
        exp_path += ".gif"

    if use_rgba:
        for img in frames:
            print(img.shape)
            assert img.shape[2] == 3, f"Image must have alpha channel! But has: {img.shape}"

        pil_frames = [_Image.fromarray(fr).convert("RGBA") for fr in frames]

        for pil_fr, fr in zip(pil_frames, frames):
            alpha_pil = _Image.fromarray(fr[:, :, 3])
            pil_fr.putalpha(alpha_pil)

    else:
        pil_frames = [_Image.fromarray(fr).convert("RGB") for fr in frames]

    pil_frames[0].save(
        exp_path, save_all=True, append_images=pil_frames[1:],
        optimize=False, loop=0,
        # background=(0, 0, 0, 255),
        quality=quality, duration=duration,
        disposal=disposal,
    )
    return 0


def stack_images_to_grid(
        images, labels=None, font_scale=1.2, font_thickness=3,
        border=2,
        auto_convert_monochannel=True,
        pics_in_row=3
):
    """
    Convert list of images to 1 Image in grid layout.
    """
    h, w, c = images[0].shape

    if labels is None:
        labels = [''] * len(images)
    elif len(labels) != len(images):
        raise ValueError("Lables size does not match images number")

    output_pic = None
    row_pic = None

    for ind, (ch, lb) in enumerate(zip(images, labels)):
        if len(ch.shape) == 2 and auto_convert_monochannel:
            ch = ch[:, :, _np.newaxis]
            rgb = _np.concatenate([ch, ch, ch], axis=2)

        elif ch.shape[2] == 4:
            rgb = ch[:, :, :3]

        elif ch.shape[2] == 1:
            rgb = _np.concatenate([ch, ch, ch], axis=2)

        else:
            # print("WHAT? What i missed?")
            # print(ch.shape)
            rgb = ch

        rgb = rgb.astype(_np.uint8)

        rgb = _cv2.putText(
            rgb, lb, (5, 30),
            fontScale=font_scale, fontFace=_cv2.FONT_HERSHEY_SIMPLEX, color=(
                50, 50, 0),
            thickness=font_thickness + 5,
        )
        rgb = _cv2.putText(
            rgb, lb, (5, 30),
            fontScale=font_scale, fontFace=_cv2.FONT_HERSHEY_SIMPLEX, color=(
                50, 255, 0),
            thickness=font_thickness,
        )

        # print(f"row: {row_pic.shape}, rgb: {rgb.shape}")
        if row_pic is None:
            row_pic = rgb
        else:
            row_pic = _np.concatenate([row_pic, rgb], 1)

        # print(f"Row Pic: {row_pic.shape}")

        if not ((ind + 1) % pics_in_row) and ind != 0:
            if output_pic is None:
                output_pic = row_pic
            else:
                output_pic = _np.concatenate([output_pic, row_pic], axis=0)
            row_pic = None

    if row_pic is not None:
        if output_pic is None:
            output_pic = row_pic
        else:
            _, match_w, _ = output_pic.shape
            blank = _np.zeros((h, match_w, c), dtype=_np.uint8)
            cur_h, cur_w, _ = row_pic.shape
            blank[:cur_h, :cur_w] = row_pic
            output_pic = _np.concatenate([output_pic, blank], axis=0)

    return output_pic


__all__ = [
    'read_webp_frames', 'read_gif_frames', 'save_image_list_to_gif',
    'stack_images_to_grid']

if __name__ == "__main__":
    import os
    img = _cv2.imread(os.path.join(os.path.dirname(__file__), "cat.jpg"))



import numpy as _np
import cv2 as _cv2

import matplotlib.pyplot as _plt


# from yasiu_image.image import stack_images_to_grid


def image_to_features(image, include_pos=False, norm_pos=True, weight_pos=1.0):
    """
    Convert each pixel into separate feature
    include_pos: bool, to add pixel location or not.
    norm_pos: bool, to normalize position to <0,1> range
    weight_pos: scale normalized position with weight
    """
    h, w, *_ = image.shape
    fts = image.reshape(h * w, -1) / 255

    if include_pos:
        y, x = _np.ogrid[:h, :w]
        XX, YY = _np.meshgrid(x, y)
        if norm_pos:
            XX = XX / (w - 1) * weight_pos
            YY = YY / (h - 1) * weight_pos

        # print(XX)
        # print(YY)
        poses = _np.stack([XX, YY], axis=-1)
        poses = poses.reshape(h * w, 2)

        fts = _np.concatenate([fts, poses], axis=1)

    return fts


__all__ = [
    'image_to_features',
    'features_to_image',
]


def features_to_image(features, shape, remove_pos_columns=False, undo_normalization=True):
    """
    Revert features to image.
    """
    if remove_pos_columns:
        features = features[:, :-2]

    if len(shape) == 3:
        h, w, c = shape
        # shape = h, w, c
    else:
        h, w = shape
        # shape = h, w

    image = features.reshape(shape)

    if undo_normalization:
        image = _np.round(image * 255).astype(_np.uint8)

    return image


if __name__ == "__main__":
    import os
    image = _cv2.imread(os.path.join(os.path.dirname(__file__), "cat.jpg"))
    # image = _cv2.resize(image, (400, 400), )

    # image = np.arange(30, dtype=np.uint8)
    # image = image.reshape((5, 6))

    "Get Features"
    fts = image_to_features(image, include_pos=True)
    # print(image)
    # print("Features:")
    # print(fts)

    # "Revert features"
    rev_img = features_to_image(fts[:, :3], image.shape)
    # cv2.imshow("Rev", image)
    # cv2.waitKey()
    # print(image)
    # print(rev_img)
    # print(image.shape, rev_img.shape)

    # stack = stack_images_to_grid([image, rev_img])
    # image = image[:, :, [2, 1, 0]]
    stack = _np.concatenate([image, rev_img], axis=1).astype(_np.uint8)
    _plt.imshow(stack)
    # plt.figure()
    _plt.colorbar()
    _plt.show()



from yasiu_image.features import image_to_features, features_to_image
import pytest
import cv2


loaded_image = cv2.imread("cat.jpg", 1)


@pytest.fixture()
def cat_full_shape():
    return loaded_image.copy()


@pytest.fixture()
def cat_gray():
    img = cv2.cvtColor(loaded_image, cv2.COLOR_BGR2GRAY)
    return img


@pytest.fixture()
def cat_gray_3d():
    img = cv2.cvtColor(loaded_image, cv2.COLOR_BGR2GRAY)
    img = img.reshape(*img.shape, 1)
    return img


@pytest.mark.parametrize('fix_name', ['cat_full_shape', 'cat_gray', 'cat_gray_3d'])
def test1_call(fix_name, request):
    cat_pic = request.getfixturevalue(fix_name)
    image_to_features(cat_pic)


@pytest.mark.parametrize('fix_name', ['cat_full_shape', 'cat_gray', 'cat_gray_3d'])
def test1_call_b(fix_name, request):
    cat_pic = request.getfixturevalue(fix_name)
    ft = image_to_features(cat_pic)
    ret = features_to_image(ft, cat_pic.shape)


@pytest.mark.parametrize('fix_name', ['cat_full_shape', 'cat_gray', 'cat_gray_3d'])
def test2_equal(fix_name, request):
    cat_pic = request.getfixturevalue(fix_name)
    ft = image_to_features(cat_pic)
    ret = features_to_image(ft, cat_pic.shape)

    assert (cat_pic == ret).all(), "Image does not match!"
    assert (cat_pic.shape == ret.shape), "Shape does not match!"


@pytest.mark.parametrize('fix_name', ['cat_full_shape', 'cat_gray', 'cat_gray_3d'])
def test3_pos_values(fix_name, request):
    cat_pic = request.getfixturevalue(fix_name)
    ft = image_to_features(cat_pic, include_pos=True)

    assert ft.shape[1] >= 3, "Features shape is less than 3 (x,y, color)"

    assert ft[0, -2] == 0, "First feat x should be 0"
    assert ft[0, -1] == 0, "First feat y should be 0"

    assert ft[-1, -2] == 1, "Last feat, x shoudl be 1"
    assert ft[-1, -1] == 1, "Last feat, y should be 1"



from setuptools import setup
from yasiu_image import VERSION


project_urls = {
        "1. Native Package": "https://pypi.org/project/yasiu-native/",
        "2. Math Package": "https://pypi.org/project/yasiu-math/",
        "3. Image Package": "https://pypi.org/project/yasiu-image/",
        "4. Visualisation Package": "https://pypi.org/project/yasiu-vis/",

        "5. Source repo": "https://github.com/GrzegorzKrug/yasiu-image",
}

classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",

        "Topic :: Multimedia :: Video",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Mathematics",

        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
]

keywords = [
        "opencv-python",
        "opencv",
        "cv2",
        "pillow",
        "numpy",
        "scipy",
        "image processing",
        "imageprocessing",
]

author = "Grzegorz Krug"
author_email = "kruggrzegorz@gmail.com"

description = "Easier image processing."
readme_path = "README.md"

with open(readme_path, "rt") as file:
    long_description = file.read()

python_requires = '>=3.7'
install_requires = [
        "opencv-python", "numpy", "Pillow",
]

setup(
        name='yasiu-image',
        version=".".join([str(num) for num in VERSION]),
        author=author,
        author_email=author_email,
        maintainer=author,
        maintainer_email=author_email,

        license='MIT',

        description=description,
        long_description=long_description,
        long_description_content_type='text/markdown',
        project_urls=project_urls,

        keywords=keywords,
        classifiers=classifiers,

        python_requires=python_requires,
        install_requires=install_requires,
        # setup_requires=install_requires,

)



import cv2 as _cv2
import numpy as _np


def resize_with_aspect_ratio(orig, resize_key='outer', new_dim=150):
    """

    :param sequence:
    :param resize_key:
        - outer
        - inner
        - height
        - width

    :param new_dim:
    :return:
    """
    h, w, c = orig.shape

    h_w_ratio = h / w
    w_h_ratio = w / h

    if h >= w and resize_key == 'outer':
        new_h = new_dim
        new_w = new_dim * w_h_ratio

    elif w >= h and resize_key == 'outer':
        new_w = new_dim
        new_h = new_dim * h_w_ratio

    elif resize_key == "height":
        new_h = new_dim
        new_w = new_dim * w_h_ratio

    elif resize_key == "width":
        new_w = new_dim
        new_h = new_dim * h_w_ratio

    elif h < w and resize_key == 'inner':
        new_h = new_dim
        new_w = new_dim * w_h_ratio
    else:
        new_w = new_dim
        new_h = new_dim * h_w_ratio

    new_h = _np.round(new_h).astype(int)
    new_w = _np.round(new_w).astype(int)

    # print(f"Resize: kwarg: {kwarg}")
    # sequence = [imutils.resize(fr, **kwarg) for fr in sequence]
    ret = _cv2.resize(orig, (new_w, new_h))
    return ret


def downscale_with_aspect_ratio(orig, resize_key='outer', max_dimension=150):
    """
    Scales picture down if size exceeds give dimension.


    :param orig:


    :param resize_key: specify which axis to resize
    :type resize_key: float

    resize_key:
        #. `outer` : resize bigger axis
        #. `inner` : resize smaller axis
        #. `height`: check height size
        #. `width` : check width size
    :type: str


    :type resize_key: str

    :param max_dimension:
    :return:
    """
    h, w, c = orig.shape

    h_w_ratio = h / w
    w_h_ratio = w / h

    if h >= w and resize_key == 'outer' and h > max_dimension:
        new_h = max_dimension
        new_w = max_dimension * w_h_ratio

    elif w >= h and resize_key == 'outer' and w > max_dimension:
        new_w = max_dimension
        new_h = max_dimension * h_w_ratio

    elif resize_key == "height" and h > max_dimension:
        new_h = max_dimension
        new_w = max_dimension * w_h_ratio

    elif resize_key == "width" and w > max_dimension:
        new_w = max_dimension
        new_h = max_dimension * h_w_ratio

    elif h < w and resize_key == 'inner' and h < max_dimension:
        new_h = max_dimension
        new_w = max_dimension * w_h_ratio
    elif w < h and resize_key == 'inner' and w < max_dimension:
        new_w = max_dimension
        new_h = max_dimension * h_w_ratio
    else:
        return orig

    new_h = _np.round(new_h).astype(int)
    new_w = _np.round(new_w).astype(int)

    # print(f"Resize: kwarg: {kwarg}")
    # sequence = [imutils.resize(fr, **kwarg) for fr in sequence]
    if new_w < w or new_h < h:
        ret = _cv2.resize(orig, (new_w, new_h))
        return ret
    return orig


def squerify(img: _np.ndarray, /, offset: float = 0.0, *, type_="clip"):
    """
    Clip excessive fraction of image.

    :param img: 2d np.ndarry

    :param type_: Only 'clip' is supported
    
    :param offset_val: float
        Pan clip to side.
        Horizonatal: -1 is left, 1 is right.
        Vertical: -1 is up, 1 is down.

        Offset = 0 clips at center.
    
    :return:
    """

    is_3d = len(img.shape) == 3

    if is_3d:
        H, W, C = img.shape
    else:
        H, W = img.shape
        C = None

    if H == W:
        "Square already"
        return img

    if type_ == "clip":
        gap = abs(H - W)
        pos = (_np.clip(offset, -1, 1) + 1) / 2

        first = (gap * pos).round().astype(int)
        second = first - gap
        if second == 0:
            second = None

        if H > W:
            new_img = img[first:second, :]
        else:
            new_img = img[:, first:second]

        # print("new shape", new_img.shape)
        return new_img
    else:
        raise KeyError("Only clip supported")


__all__ = [
    'squerify',
    'resize_with_aspect_ratio',
    'downscale_with_aspect_ratio',
]


if __name__ == "__main__":
    import os
    img = _cv2.imread(os.path.join(os.path.dirname(__file__), "cat.jpg"))
    smol = downscale_with_aspect_ratio(img, max_dimension=400)
    smol2 = downscale_with_aspect_ratio(smol, max_dimension=600)
    img = smol[200:, :, :]

    _cv2.imshow("Orig", img)
    # _cv2.imshow("Smol", smol)
    # _cv2.imshow("Smol600", smol2)

    sq = squerify(img, -1)
    _cv2.imshow("Cat Square -1", sq)
    sq = squerify(img, 0)
    _cv2.imshow("Cat Square 0 ", sq)
    sq = squerify(img, 0.5)
    _cv2.imshow("Cat Square 1", sq)
    _cv2.waitKey()
